/* 
   NSAccount.m

   Copyright (C) 1995, 1996 Ovidiu Predescu and Mircea Oancea.
   All rights reserved.

   Author: Mircea Oancea <mircea@jupiter.elcom.pub.ro>

   This file is part of libFoundation.

   Permission to use, copy, modify, and distribute this software and its
   documentation for any purpose and without fee is hereby granted, provided
   that the above copyright notice appear in all copies and that both that
   copyright notice and this permission notice appear in supporting
   documentation.

   We disclaim all warranties with regard to this software, including all
   implied warranties of merchantability and fitness, in no event shall
   we be liable for any special, indirect or consequential damages or any
   damages whatsoever resulting from loss of use, data or profits, whether in
   an action of contract, negligence or other tortious action, arising out of
   or in connection with the use or performance of this software.
*/

#include <Foundation/common.h>

#include <sys/types.h>

#if HAVE_PWD_H
#include <pwd.h>
#endif

#if HAVE_GRP_H
#include <grp.h>
#endif

#ifdef HAVE_LIBC_H
#  include <libc.h>
#else
#  include <unistd.h>
#endif

#if HAVE_WINDOWS_H
# include <windows.h>
#endif

#ifdef __WIN32__
struct passwd {
    const char *pw_name;
    int pw_uid;
};
#endif

#include <Foundation/NSString.h>
#include <Foundation/NSArray.h>
#include <Foundation/NSUtilities.h>
#include <Foundation/NSLock.h>

#include "NSAccount.h"

extern NSRecursiveLock* libFoundationLock;

@implementation NSAccount

// Creating an account

+ currentAccount 
{
    [self subclassResponsibility:_cmd];
    return nil; 
}

+ accountWithName:(NSString*)name
{
    [self subclassResponsibility:_cmd];
    return nil; 
}

+ accountWithNumber:(unsigned int)number
{
    [self subclassResponsibility:_cmd];
    return nil; 
}

// Getting account information

- (NSString*)accountName
{
    [self subclassResponsibility:_cmd];
    return nil; 
}

- (unsigned)accountNumber
{
    [self subclassResponsibility:_cmd];
    return 0; 
}

// Copying Protocol

- copy
{
    return RETAIN(self);
}

- copyWithZone:(NSZone*)zone
{
    return RETAIN(self);
}

@end /* NSAccount */

/*
 *  User Account
 */

@implementation NSUserAccount

// Init & dealloc

- initWithPasswordStructure:(struct passwd*)ptr
{
    name = RETAIN([NSString stringWithCStringNoCopy:Strdup(ptr->pw_name)
                            freeWhenDone:YES]);
    fullName = RETAIN([NSString stringWithCStringNoCopy:Strdup(ptr->pw_gecos)
                                freeWhenDone:YES]);
    homeDirectory = RETAIN([NSString stringWithCStringNoCopy:Strdup(ptr->pw_dir)
                                     freeWhenDone:YES]);
    userNumber = ptr->pw_uid;

    return self;
}

- (void)dealloc
{
    RELEASE(name);
    RELEASE(fullName);
    RELEASE(homeDirectory);
    [super dealloc];
}

// Creating an account

static NSUserAccount* currentUserAccount = nil;

+ (void)initialize
{
    static BOOL initialized = NO;

    if (!initialized) {
	[libFoundationLock lock];

#ifndef __WIN32__
	/* Initialize the group account global variable */
	[NSGroupAccount initialize];
#endif

#if HAVE_GETUID
    {
	struct passwd* ptr = getpwuid(getuid());

	currentUserAccount = [self alloc];
	[currentUserAccount initWithPasswordStructure:ptr];
    }
#else
    {
	char n[60];
	int len = 60;
	if (GetUserName (n, &len))
	    currentUserAccount
                = RETAIN([self accountWithName:[NSString stringWithCString:n]]);
        
    }
#endif
	[libFoundationLock unlock];
	initialized = YES;
    }
}

+ currentAccount 
{
    return currentUserAccount;
}

+ accountWithName:(NSString*)aName
{
#if HAVE_GETPWNAM
    struct passwd* ptr;

    [libFoundationLock lock];
    ptr = getpwnam((char*)[aName cString]);
    [libFoundationLock unlock];
    return ptr
        ? AUTORELEASE([[self alloc] initWithPasswordStructure:ptr])
	: nil;
#else
    struct passwd pt;
    pt.pw_name = [aName cString];
    pt.pw_uid = 0;
    return AUTORELEASE([[self alloc] initWithPasswordStructure:&pt]);
#endif
}

+ accountWithNumber:(unsigned int)number
{
#if HAVE_GETPWUID
    struct passwd* ptr;

    [libFoundationLock lock];
    ptr = getpwuid(number);
    [libFoundationLock unlock];

    return ptr
        ? AUTORELEASE([[self alloc] initWithPasswordStructure:ptr])
	: nil;
#else
    return 0;
#endif
}

- (NSString*)accountName	{return name;}
- (unsigned)accountNumber	{return userNumber;}
- (NSString*)fullName		{return fullName;}
- (NSString*)homeDirectory	{return homeDirectory;}

@end /* NSUserAccount */

#ifndef __WIN32__
/*
 *  Group Account
 */

@implementation NSGroupAccount

// Init & dealloc

- initWithGroupStructure:(struct group*)ptr
{
    int cnt;

    name = RETAIN([NSString stringWithCString:ptr->gr_name]);
    groupNumber = ptr->gr_gid;

    // count group members
    for (cnt = 0; ptr->gr_mem[cnt]; cnt++)
	;
    
    {
	NSString** array = Malloc (cnt * sizeof(id));
	int i;
	
	for (i = 0; i < cnt; i++)
	    array[i] = [NSString stringWithCString:ptr->gr_mem[i]];
	members = [[NSArray alloc] initWithObjects:array count:cnt];
        Free (array);
    }
    
    return self;
}

- (void)dealloc
{
    RELEASE(name);
    RELEASE(members);
    [super dealloc];
}

// Creating an account

static NSGroupAccount* currentGroupAccount = nil;

+ (void)initialize
{
    static BOOL initialized = NO;

    if (!initialized) {
	[libFoundationLock lock];
	currentGroupAccount = RETAIN([self accountWithNumber:getgid()]);
	[libFoundationLock unlock];
	initialized = YES;
    }
}

+ currentAccount 
{
    return currentGroupAccount;
}

+ accountWithName:(NSString*)aName
{
    struct group* ptr;

    [libFoundationLock lock];
    ptr = getgrnam((char*)[aName cString]);
    [libFoundationLock unlock];

    return ptr
        ? AUTORELEASE([[self alloc] initWithGroupStructure:ptr])
	: nil;
}

+ accountWithNumber:(unsigned int)number
{
    struct group* ptr;

    [libFoundationLock lock];
    ptr = getgrgid(number);
    [libFoundationLock unlock];

    return ptr
        ? AUTORELEASE([[self alloc] initWithGroupStructure:ptr])
	: nil;
}

- (NSString*)accountName	{return name;}
- (unsigned)accountNumber	{return groupNumber;}
- (NSArray*)members		{return members;}

@end /* NSGroupAccount */
#endif
/*
  Local Variables:
  c-basic-offset: 4
  tab-width: 8
  End:
*/

