/* 
   NSBundle.m

   Copyright (C) 1995, 1996 Ovidiu Predescu and Mircea Oancea.
   All rights reserved.

   Author: Mircea Oancea <mircea@jupiter.elcom.pub.ro>

   This file is part of libFoundation.

   Permission to use, copy, modify, and distribute this software and its
   documentation for any purpose and without fee is hereby granted, provided
   that the above copyright notice appear in all copies and that both that
   copyright notice and this permission notice appear in supporting
   documentation.

   We disclaim all warranties with regard to this software, including all
   implied warranties of merchantability and fitness, in no event shall
   we be liable for any special, indirect or consequential damages or any
   damages whatsoever resulting from loss of use, data or profits, whether in
   an action of contract, negligence or other tortious action, arising out of
   or in connection with the use or performance of this software.
*/

#include <Foundation/common.h>
#include <Foundation/NSBundle.h>
#include <Foundation/NSArray.h>
#include <Foundation/NSDictionary.h>
#include <Foundation/NSString.h>
#include <Foundation/NSAutoreleasePool.h>
#include <Foundation/NSProcessInfo.h>
#include <Foundation/NSNotification.h>
#include <Foundation/NSUserDefaults.h>
#include <Foundation/NSFileManager.h>
#include <extensions/objc-runtime.h>

/*
 * Static class variables
 */

typedef struct {
    Class class;
    Category* category;
} LoadingClassCategory;

static NSMapTable*	    bundleClasses = NULL;// class -> bundle mapping
static NSMapTable*	    bundleNames = NULL;	// path  -> bundle mapping
static NSBundle*	    mainBundle = nil;	// application bundle
static LoadingClassCategory*load_Classes = NULL;// used while loading
static int		    load_classes_size = 0;// used while loading
static int		    load_classes_used = 0;// used while loading

/*
 * NSBundle methods
 */

@implementation NSBundle

// Library resource directory

+ (NSString*)_fileResourceNamed:(NSString*)fileName
  extension:(NSString*)extension
  inDirectory:(NSString*)directoryName
{
    NSDictionary* environment = [[NSProcessInfo processInfo] environment];
    NSString* resourcesPathsString;
    NSMutableArray* resourcesPaths;
    NSString* fullFilenamePath = nil;
    int i, count;
#ifndef WITH_GNUSTEP
    NSArray* searchPaths = [NSArray arrayWithObject:@RESOURCES_PATH];
#else
    NSArray* searchPaths = [NSArray arrayWithObjects:
	[[[[environment objectForKey:@"GNUSTEP_USER_ROOT"]
	      stringByAppendingPathComponent:@"Libraries"]
	      stringByAppendingPathComponent:@"Resources"]
	      stringByAppendingPathComponent:@"libFoundation"],
	[[[[environment objectForKey:@"GNUSTEP_LOCAL_ROOT"]
	      stringByAppendingPathComponent:@"Libraries"]
	      stringByAppendingPathComponent:@"Resources"]
	      stringByAppendingPathComponent:@"libFoundation"],
	[[[[environment objectForKey:@"GNUSTEP_SYSTEM_ROOT"]
	      stringByAppendingPathComponent:@"Libraries"]
	      stringByAppendingPathComponent:@"Resources"]
	      stringByAppendingPathComponent:@"libFoundation"],
	nil];
#endif

    resourcesPathsString
	= [environment objectForKey:@"LIB_FOUNDATION_RESOURCES_PATH"];
    if (resourcesPathsString) {
	resourcesPaths
	    = AUTORELEASE([[resourcesPathsString
                               componentsSeparatedByString:@":"]
                              mutableCopy]);
	[resourcesPaths addObjectsFromArray:searchPaths];
	searchPaths = resourcesPaths;
    }

    for (i = 0, count = [searchPaths count]; i < count; i++) {
	fullFilenamePath = [[[[[searchPaths objectAtIndex:i]
				  stringByAppendingPathComponent:directoryName]
				  stringByAppendingPathComponent:fileName]
				  stringByAppendingPathExtension:extension] 
				  stringByResolvingSymlinksInPath];
	if(fullFilenamePath && [fullFilenamePath length])
	    break;
    }

    return fullFilenamePath;
}

// Bundle initialization

+ (void)initialize
{
    mainBundle = nil;
    bundleClasses = NSCreateMapTable(NSNonOwnedPointerMapKeyCallBacks,
	NSNonRetainedObjectMapValueCallBacks, 23);
    bundleNames = NSCreateMapTable(NSObjectMapKeyCallBacks,
	NSNonRetainedObjectMapValueCallBacks, 23);
}

// Load info for bundle

- (void)loadInfo
{
    NSString* file;

    if (infoDictionary)
	return;
    
#if WITH_GNUSTEP
    file = [self pathForResource:@"Info-gnustep" ofType:@"plist"];
#else
    file = [self pathForResource:@"Info" ofType:@"plist"];
#endif
    
    if (file)
	infoDictionary = RETAIN([[NSString stringWithContentsOfFile:file] 
                                    propertyList]);
    
    if (!infoDictionary)
	infoDictionary = [[NSDictionary alloc] init];
}

// Internal code loading

static void load_callback(Class class, Category* category)
{
    if (load_classes_used >= load_classes_size) {
	load_classes_size += 128;
	load_Classes = Realloc(load_Classes,
	    load_classes_size*sizeof(LoadingClassCategory));
    }
    load_Classes[load_classes_used].class = class;
    load_Classes[load_classes_used].category = category;
    load_classes_used++;
}

- (BOOL)loadCode
{
    int i;
    NSString* file;
    NSString* rfile;
    BOOL status;
    int objc_load_module(const char*, void (*)(Class, Category*));
#if WITH_GNUSTEP
    NSDictionary* environment = [[NSProcessInfo processInfo] environment];
#endif

    if (codeLoaded)
	return YES;
    else
	codeLoaded = YES;
    
    // Find file to load
    if ((file = [[self infoDictionary] objectForKey:@"NSExecutable"]))
#if WITH_GNUSTEP
      file = [[[[fullPath stringByAppendingPathComponent:
			    [environment objectForKey:@"GNUSTEP_HOST_CPU"]]
		 stringByAppendingPathComponent:
		   [environment objectForKey:@"GNUSTEP_HOST_OS"]]
		stringByAppendingPathComponent:
		  [environment objectForKey:@"LIBRARY_COMBO"]]
	       stringByAppendingPathComponent:file];
#else
	file = [fullPath stringByAppendingPathComponent:file];
#endif
    else
#if WITH_GNUSTEP
	file = [[[[fullPath stringByAppendingPathComponent:
			      [environment objectForKey:@"GNUSTEP_HOST_CPU"]]
		   stringByAppendingPathComponent:
		     [environment objectForKey:@"GNUSTEP_HOST_OS"]]
		  stringByAppendingPathComponent:
		    [environment objectForKey:@"LIBRARY_COMBO"]]
		  stringByAppendingPathComponent:
		   [[fullPath lastPathComponent] stringByDeletingPathExtension]];
#else
	file = [fullPath stringByAppendingPathComponent:
	    [[fullPath lastPathComponent] stringByDeletingPathExtension]];
#endif
    
    rfile = [file stringByResolvingSymlinksInPath];
    if (!rfile) {
	NSLog(@"%@:NSBundle: cannot find executable file %@",
	    [[NSProcessInfo processInfo] processName], file);
	return NO;
    }
    
    // Prepare to keep classes/categories loaded
    load_classes_size = 128;
    load_classes_used = 0;
    load_Classes = MallocAtomic(load_classes_size
				* sizeof(LoadingClassCategory));

    status = objc_load_module([file fileSystemRepresentation], load_callback);
    
    if (status) {
	NSDictionary* info;
	firstLoadedClass = Nil;
	for (i = 0; i < load_classes_used; i++) {
	    // get first class from bundle
	    if (!firstLoadedClass && !load_Classes[i].category)
		firstLoadedClass = load_Classes[i].class;
	    
	    // TODO - call class/category load method
	    
	    // insert in bundle hash
	    if (!load_Classes[i].category)
		NSMapInsert(bundleClasses, load_Classes[i].class, self);
	    
	    // post notification
	    if (!load_Classes[i].category)
		info = [NSDictionary dictionaryWithObjectsAndKeys:
			    [NSString stringWithCStringNoCopy:
			    (char*)(load_Classes[i].class->name)
			    freeWhenDone:NO], 
			@"NSClass", 
			    nil];
	    else
		info = [NSDictionary dictionaryWithObjectsAndKeys:
			    [NSString stringWithCStringNoCopy:
			    (char*)(load_Classes[i].class->name)
			    freeWhenDone:NO],
			@"NSClass",
			    [NSString stringWithCStringNoCopy:
			    (char*)(load_Classes[i].category->category_name)
			    freeWhenDone:NO], 
			@"NSCategory",
			     nil];
	    [[NSNotificationCenter defaultCenter] 
		postNotificationName:@"NSBundleDidLoadNotification"
		object:nil
		userInfo:info];
	}
        [[NSNotificationCenter defaultCenter] 
	    postNotificationName:@"NSBundleDidLoadNotification"
	    object:nil
	    userInfo:nil];
    }
    
    Free(load_Classes); load_Classes = NULL;
    
    return status;
}

// Initializing an NSBundle 

static BOOL canReadDirectory(NSString* path)
{
    NSFileManager* fileManager = [NSFileManager defaultManager];
    BOOL isDirectory;

    if (![fileManager fileExistsAtPath:path isDirectory:&isDirectory]
	|| !isDirectory)
	return NO;

    return [fileManager isReadableFileAtPath:path];
}

static BOOL canReadFile(NSString* path)
{
    NSFileManager* fileManager = [NSFileManager defaultManager];

    return [fileManager isReadableFileAtPath:path];
}

- (id)initWithPath:(NSString*)path
{
    NSBundle* old;

    path = [path stringByResolvingSymlinksInPath];
    
    if (!path || !canReadDirectory(path)) {
	RELEASE(self);
	return nil;
    }
    
    old = (NSBundle*)NSMapGet(bundleNames, path);
    if (old) {
	(void)AUTORELEASE(self);
	return RETAIN(old);
    }
    
    NSMapInsert(bundleNames, path, self);
    fullPath = RETAIN(path);
    return self;
}

// TODO - now bundle is not capable of dealloc & code unloading

- retain {return self;}
- autorelease {return self;}
- (void)release{}
- (unsigned int)retainCount {return 1;}

- (void)dealloc
{
    NSMapRemove(bundleNames, fullPath);
    RELEASE(fullPath);
    RELEASE(infoDictionary);
    RELEASE(stringTables);
    [super dealloc];
}

// Getting an NSBundle 

+ (NSBundle*)bundleForClass:(Class)aClass
{
    NSBundle* bundle = (NSBundle*)NSMapGet(bundleClasses, aClass);
    
    return bundle ? bundle : [self mainBundle];
}

+ (NSBundle*)bundleWithPath:(NSString*)path
{
    return AUTORELEASE([[self alloc] initWithPath:path]);
}

+ (NSBundle*)mainBundle
{
    if (!mainBundle) {
	NSString* path = [[[[NSProcessInfo processInfo] arguments]
			      objectAtIndex:0]
			      stringByDeletingLastPathComponent];
	if ([path isEqual:@""])
	    path = @".";
#if WITH_GNUSTEP
	else {
	    /* The path is the complete path to the executable, including the
	       processor, the OS and the library combo. Strip these directories
	       from the main bundle's path. */
	    path = [[[path stringByDeletingLastPathComponent]
			    stringByDeletingLastPathComponent]
			    stringByDeletingLastPathComponent];
	}
#endif
	mainBundle = [[NSBundle alloc] initWithPath:path];
    }
    return mainBundle;
}

// Getting a Bundled Class 

- (Class)classNamed:(NSString*)className
{
    Class class;

    [self loadCode];
    
    class = NSClassFromString(className);
    if (class && (NSBundle*)NSMapGet(bundleClasses, class) == self)
	return class;

    return nil;
}

- (Class)principalClass
{
    NSString* className;
    Class class;
    
    [self loadCode];
    className = [[self infoDictionary] objectForKey:@"NSPrincipalClass"];
    
    class = NSClassFromString(className);
    
    return class ? class : firstLoadedClass;
}

// Finding a Resource 

+ (NSString*)pathForResource:(NSString*)name ofType:(NSString*)ext
  inDirectories:(NSArray*)directories
{
    int i, n;
    NSString* file;
    CREATE_AUTORELEASE_POOL(pool);

    if (ext)
	name = [name stringByAppendingPathExtension:ext];
    
    n = [directories count];
    
    for (i = 0; i < n; i++) {
	file = [[directories objectAtIndex:i]
	    stringByAppendingPathComponent:name];
	if (canReadFile(file))
	    goto found;
    }
    
    file = nil;
    
    found:
    
    (void)RETAIN(file);
    RELEASE(pool);
    
    return AUTORELEASE(file);
}

- (NSString*)pathForResource:(NSString*)name ofType:(NSString*)ext
 inDirectory:(NSString*)directory
{
    int i, n;
    NSString* path;
    NSString* file;
    NSMutableArray* languages;
    CREATE_AUTORELEASE_POOL(pool);
    
    // Get languages and translate list by adding "lproj" extension
    // {English, German, ...} to {English.lproj, German.lproj, ...}
    languages = AUTORELEASE([[[NSUserDefaults standardUserDefaults] 
                                 stringArrayForKey:@"Languages"] mutableCopy]);
    n = [languages count];
    for (i = 0; i < n; i++) {
	file = [[languages objectAtIndex:i] 
	    stringByAppendingPathExtension:@"lproj"];
	[languages replaceObjectAtIndex:i withObject:file];
    }
    
    // make file name name.ext if extension is present
    if (ext)
	name = [name stringByAppendingPathExtension:ext];
    
    // look for fullPath/Resources/directory/...
    path = [fullPath stringByAppendingPathComponent:@"Resources"];
    if (directory && ![directory isEqualToString:@""])
	path = [path stringByAppendingPathComponent:directory];
    if (canReadDirectory(path)) {
	// check languages
	for (i = 0; i < n; i++) {
	    file = [[path stringByAppendingPathComponent:
		[languages objectAtIndex:i]]
		    stringByAppendingPathComponent:name];
	    if (canReadFile(file))
		goto found;
	}
	// check base
	file = [path stringByAppendingPathComponent:name];
	if (canReadFile(file))
	    goto found;
    }
    
    // look for fullPath/directory/...
    if (directory && ![directory isEqualToString:@""])
	path = [fullPath stringByAppendingPathComponent:directory];
    else
	path = fullPath;
    if (canReadDirectory(path)) {
	// check languages
	for (i = 0; i < n; i++) {
	    file = [[path stringByAppendingPathComponent:
		[languages objectAtIndex:i]]
		    stringByAppendingPathComponent:name];
	    if (canReadFile(file))
		goto found;
	}
	// check base
	file = [path stringByAppendingPathComponent:name];
	if (canReadFile(file))
	    goto found;
    }

    file = nil;
    
found:
    
    (void)RETAIN(file);
    RELEASE(pool);
    
    return AUTORELEASE(file);
}

- (NSArray*)pathsForResourcesOfType:(NSString*)extension
  inDirectory:(NSString*)bundlePath
{
    // TODO - needs NSFileManager
    [self notImplemented:_cmd];
    return nil;
}

- (NSString*)pathForResource:(NSString*)name ofType:(NSString*)ext
{
    return [self pathForResource:name ofType:ext inDirectory:nil];
}

- (NSString*)resourcePath
{
    return [fullPath stringByAppendingPathComponent:@"Resources"];
}

// Getting bundle information

- (NSDictionary*)infoDictionary
{
    [self loadInfo];
    return infoDictionary;
}

// Getting the Bundle Directory 

- (NSString*)bundlePath
{
    return fullPath;
}

// Managing Localized Resources

- (NSString*)localizedStringForKey:(NSString*)key value:(NSString*)value
  table:(NSString*)tableName
{
    NSDictionary* table;
    NSString* string;

    if (!stringTables)
	stringTables = [NSMutableDictionary new];
    
    table = [stringTables objectForKey:tableName];
    if (!table) {
	string = [NSString stringWithContentsOfFile:
		[self pathForResource:tableName ofType:@"strings"]];
	if (!string)
		return value;
	table = [string propertyListFromStringsFileFormat];
	if (table)
	    [stringTables setObject:table forKey:tableName];
    }

    string = [table objectForKey:key];
    if (!string)
	string = value;

    return string;
}

- (void)releaseStringtableCache
{
    RELEASE(stringTables);
    stringTables = nil;
}

- (NSString*)description
{
    /* Don't use -[NSString stringWithFormat:] method because it can cause
       infinite recursion. */
    char buffer[1024];

    sprintf (buffer,
	    "<%s %p fullPath: %s infoDictionary: %p stringTables: %p",
	    (char*)object_get_class_name(self),
	    self,
	    fullPath ? [fullPath cString] : "nil",
	    infoDictionary,
	    stringTables);

    return [NSString stringWithCString:buffer];
}

@end /* NSBundle */
/*
  Local Variables:
  c-basic-offset: 4
  tab-width: 8
  End:
*/

