/* 
   NSConcreteArray.m

   Copyright (C) 1995, 1996 Ovidiu Predescu and Mircea Oancea.
   All rights reserved.

   Author: Mircea Oancea <mircea@jupiter.elcom.pub.ro>

   This file is part of libFoundation.

   Permission to use, copy, modify, and distribute this software and its
   documentation for any purpose and without fee is hereby granted, provided
   that the above copyright notice appear in all copies and that both that
   copyright notice and this permission notice appear in supporting
   documentation.

   We disclaim all warranties with regard to this software, including all
   implied warranties of merchantability and fitness, in no event shall
   we be liable for any special, indirect or consequential damages or any
   damages whatsoever resulting from loss of use, data or profits, whether in
   an action of contract, negligence or other tortious action, arising out of
   or in connection with the use or performance of this software.
*/

#include <stdio.h>

#include <Foundation/common.h>
#include <Foundation/NSArray.h>
#include <Foundation/NSString.h>
#include <Foundation/NSException.h>
#include <Foundation/exceptions/GeneralExceptions.h>

#include <extensions/objc-runtime.h>

#include "NSConcreteArray.h"

static NSConcreteEmptyArray *sharedEmptyArray = nil;

/*
 * NSConcreteArray class
 */

@implementation NSConcreteArray

+ (void)initialize {
    if (sharedEmptyArray == nil)
        sharedEmptyArray = [[NSConcreteEmptyArray alloc] init];
}

- (id)init
{
    RELEASE(self);
    return RETAIN(sharedEmptyArray);
}

- (id)initWithObjects:(id *)objects count:(unsigned int)count
{
    unsigned i;

    if (count == 0) {
        RELEASE(self);
        return RETAIN(sharedEmptyArray);
    }
    else if (count == 1) {
        RELEASE(self);
        return [[NSConcreteSingleObjectArray alloc]
                                             initWithObjects:objects count:1];
    }

    self->items      = NSZoneCalloc([self zone], sizeof(id), count);
    self->itemsCount = count;
    for (i = 0; i < count; i++) {
	if (!(self->items[i] = RETAIN(objects[i]))) {
	    THROW([[InvalidArgumentException alloc] 
		    initWithReason:@"Nil object to be added in array"]);
        }
    }
    return self;
}

- (id)initWithArray:(NSArray *)anotherArray
{
    unsigned i, count = [anotherArray count];

    if (count == 0) {
        RELEASE(self);
        return RETAIN(sharedEmptyArray);
    }
    else if (count == 1) {
        id item = [anotherArray objectAtIndex:0];
        
        RELEASE(self);
        return [[NSConcreteSingleObjectArray alloc]
                                             initWithObjects:&item count:1];
    }
    
    self->items = NSZoneCalloc([self zone], sizeof(id), count);
    self->itemsCount = count;
    for (i = 0; i < itemsCount; i++)
	self->items[i] = RETAIN([anotherArray objectAtIndex:i]);
    return self;
}

- (void)dealloc
{
    unsigned int index;
    for (index = 0; index < itemsCount; index++)
	RELEASE(items[index]);
    Free(items);
    [super dealloc];
}

/* Querying the Array */

- (id)objectAtIndex:(unsigned int)index
{
    if (index >= itemsCount) {
	THROW([[RangeException alloc] 
		initWithReason:@"objectAtIndex: in NSArray" 
		size:itemsCount index:index]);
    }
    return items[index];
}

- (unsigned int)count
{
    return itemsCount;
}

- (unsigned int)indexOfObjectIdenticalTo:(id)anObject
{
    unsigned i;
    for (i=0; i < itemsCount; i++)
	if (items[i] == anObject)
		return i;
    return NSNotFound;
}

@end

/*
 * NSConcreteEmptyArray class
 */

@implementation NSConcreteEmptyArray

- (id)init
{
    return self;
}
- (id)initWithObjects:(id *)objects count:(unsigned int)count
{
    if (count > 0) {
        RELEASE(self);
        self = [[NSConcreteArray alloc] initWithObjects:objects count:count];
    }
    return self;
}
- (id)initWithArray:(NSArray *)anotherArray
{
    if ([anotherArray count] > 0) {
        AUTORELEASE(self);
        self = [[NSConcreteArray alloc] initWithArray:anotherArray];
    }
    return self;
}

- (id)objectAtIndex:(unsigned int)index
{
    THROW([[RangeException alloc] 
              initWithReason:@"objectAtIndex: in NSArray" size:0 index:index]);
    return nil;
}
- (unsigned int)count
{
    return 0;
}
- (unsigned int)indexOfObjectIdenticalTo:(id)anObject
{
    return NSNotFound;
}

@end

/*
 * NSConcreteSingleObjectArray class
 */

@implementation NSConcreteSingleObjectArray

+ (void)initialize {
    // force setup of shared emtpy
    if (sharedEmptyArray == nil)
        sharedEmptyArray = [[NSConcreteEmptyArray alloc] init];
}

- (id)init
{
    RELEASE(self);
    return RETAIN(sharedEmptyArray);
}
- (id)initWithObjects:(id *)objects count:(unsigned int)count
{
    if (count == 0) {
        RELEASE(self);
        return RETAIN(sharedEmptyArray);
    }
    if (count > 1) {
        RELEASE(self);
        self = [[NSConcreteArray allocWithZone:[self zone]]
                                 initWithObjects:objects count:count];
    }

    self->item = RETAIN(objects[0]);

    return self;
}
- (id)initWithArray:(NSArray *)anotherArray
{
    unsigned count = [anotherArray count];
    
    if (count == 0) {
        RELEASE(self);
        return RETAIN(sharedEmptyArray);
    }
    if (count > 1) {
        RELEASE(self);
        self = [[NSConcreteArray alloc] initWithArray:anotherArray];
    }
    
    self->item = RETAIN([anotherArray objectAtIndex:0]);
    return self;
}

- (void)dealloc
{
    RELEASE(self->item);
    [super dealloc];
}

// query array

- (id)objectAtIndex:(unsigned int)index
{
    if (index > 0) {
      THROW([[RangeException alloc] 
                initWithReason:@"objectAtIndex: in NSArray" size:1 index:index]);
    }
    return item;
}
- (unsigned int)count
{
    return 1;
}
- (unsigned int)indexOfObjectIdenticalTo:(id)anObject
{
    return (self->item == anObject) ? 0 : NSNotFound;
}

@end

/*
 * NSConcreteMutableArray class
 */

@implementation NSConcreteMutableArray

+ (void)initialize
{
    static BOOL initialized = NO;
    if(!initialized) {
	initialized = YES;
	class_add_behavior(self, [NSConcreteArray class]);
    }
}

- (id)init
{
    self->items      = NSZoneCalloc([self zone], 1, sizeof(id));
    self->maxItems   = 1;
    self->itemsCount = 0;
    return self;
}

- (id)initWithCapacity:(unsigned int)aNumItems
{
    self->items      = NSZoneCalloc([self zone], sizeof(id), aNumItems);
    self->maxItems   = aNumItems;
    self->itemsCount = 0;
    return self;
}

- (id)initWithObjects:(id *)objects count:(unsigned int)count
{
    unsigned i;

    items = NSZoneCalloc([self zone], sizeof(id), count);
    maxItems = itemsCount = count;

    for (i=0; i < count; i++)
	if (!(items[i] = RETAIN(objects[i])))
	    THROW([[InvalidArgumentException alloc] 
		    initWithReason:@"Nil object to be added in array"]);
    return self;
}

- (id)initWithArray:(NSArray *)anotherArray
{
    unsigned i, count = [anotherArray count];

    items = NSZoneCalloc([self zone], sizeof(id), count);
    maxItems = itemsCount = count;
    for (i = 0; i < itemsCount; i++)
	items[i] = RETAIN([anotherArray objectAtIndex:i]);
    return self;
}

- (void)dealloc
{
    unsigned int index;
    for (index = 0; index < itemsCount; index++)
	RELEASE(items[index]);
    Free(items);
    [super dealloc];
}

/* Altering the Array */

- (void)insertObject:(id)anObject atIndex:(unsigned int)index
{
    unsigned int i;
    if (!anObject)
	THROW([[InvalidArgumentException alloc] 
		initWithReason:@"Nil object to be added in array"]);
    if (index > itemsCount)
	THROW([[RangeException alloc] 
		initWithReason:@"__insertObject:atIndex: in NSMutableArray" 
		size:itemsCount index:index]);
    if (itemsCount == maxItems) {
	if (maxItems) {
	    maxItems += (maxItems >> 1) ? (maxItems >>1) : 1;
	}
	else {
	    maxItems = 1;
	}
	items = (id*)Realloc(items, sizeof(id) * maxItems);
    }
    for(i = itemsCount; i > index; i--)
	items[i] = items[i - 1];
    items[index] = RETAIN(anObject);
    itemsCount++;
}

- (void)replaceObjectAtIndex:(unsigned int)index  withObject:(id)anObject
{
    if (!anObject)
	THROW([[InvalidArgumentException alloc] 
		initWithReason:@"Nil object to be added in array"]);
    if (index >= itemsCount)
	THROW([[RangeException alloc] 
		initWithReason:@"NSConcreteMutableArray replaceObjectAtIndex" 
		size:itemsCount index:index]);
    ASSIGN(items[index], anObject);
}

- (void)removeObjectsFrom:(unsigned int)index
  count:(unsigned int)count
{
    unsigned i;
    if (index + count > itemsCount)
	THROW([[RangeException alloc]
		initWithReason:@"removeObjectsFrom:count in NSMutableArray"
		size:itemsCount index:index]);
    if (!count)
	return;

#if !LIB_FOUNDATION_BOEHM_GC
    for (i = index; i < index + count; i++)
	AUTORELEASE(items[i]);
#endif

    for (i = index + count; i < itemsCount; i++, index++)
	items[index] = items[i];
    for (; index < itemsCount; index++)
	items[index] = (id)0x3;

    itemsCount -= count;
}

- (void)removeObjectsInRange:(NSRange)aRange
{
    [self removeObjectsFrom:aRange.location count:aRange.length];
}

- (void)removeAllObjects
{
    [self removeObjectsFrom:0 count:itemsCount];
}

- (void)removeLastObject
{
    if (itemsCount)
	[self removeObjectsFrom:(itemsCount - 1) count:1];
}

- (void)removeObjectAtIndex:(unsigned int)index
{
    [self removeObjectsFrom:index count:1];
}

@end
/*
  Local Variables:
  c-basic-offset: 4
  tab-width: 8
  End:
*/

