/* 
   NSConcreteDictionary.m

   Copyright (C) 1995, 1996 Ovidiu Predescu and Mircea Oancea.
   All rights reserved.

   Author: Mircea Oancea <mircea@jupiter.elcom.pub.ro>

   This file is part of libFoundation.

   Permission to use, copy, modify, and distribute this software and its
   documentation for any purpose and without fee is hereby granted, provided
   that the above copyright notice appear in all copies and that both that
   copyright notice and this permission notice appear in supporting
   documentation.

   We disclaim all warranties with regard to this software, including all
   implied warranties of merchantability and fitness, in no event shall
   we be liable for any special, indirect or consequential damages or any
   damages whatsoever resulting from loss of use, data or profits, whether in
   an action of contract, negligence or other tortious action, arising out of
   or in connection with the use or performance of this software.
*/

#include <stdarg.h>

#include <Foundation/common.h>
#include <Foundation/NSObject.h>
#include <Foundation/NSDictionary.h>
#include <Foundation/NSArray.h>
#include <Foundation/NSString.h>
#include <Foundation/NSException.h>
#include <Foundation/exceptions/GeneralExceptions.h>

#include <extensions/objc-runtime.h>

#include "NSConcreteDictionary.h"

static NSConcreteEmptyDictionary *sharedEmptyDict = nil;

/*
 * NSConcreteDictionary class
 */

@implementation NSConcreteDictionary

+ (void)initialize {
    if (sharedEmptyDict == nil)
        sharedEmptyDict = [[NSConcreteEmptyDictionary alloc] init];
}

/* Allocating and Initializing */

- (id)init
{
    RELEASE(self);
    return RETAIN(sharedEmptyDict);
}

- (id)initWithObjects:(id*)objects
    forKeys:(id*)keys 
    count:(unsigned int)count
{
    if (count == 0) {
        RELEASE(self);
        return RETAIN(sharedEmptyDict);
    }
    else if (count == 1) {
        RELEASE(self);
        return [[NSConcreteSingleObjectDictionary alloc]
                   initWithObjects:objects forKeys:keys count:count];
    }
    
    table = NSCreateMapTableWithZone(NSObjectMapKeyCallBacks,
		NSObjectMapValueCallBacks, 
		(count * 4) / 3,
		[self zone]);

    while(count--) {
	if (!keys[count] || !objects[count])
	    THROW([[InvalidArgumentException alloc] 
		    initWithReason:@"Nil object to be added in dictionary"]);
	NSMapInsert(table, keys[count], objects[count]);
    }
    return self;
}

- (id)initWithDictionary:(NSDictionary*)dictionary
{
    NSEnumerator *keys = [dictionary keyEnumerator];
    id key;

    table = NSCreateMapTableWithZone(NSObjectMapKeyCallBacks,
		NSObjectMapValueCallBacks, 
		([dictionary count] * 4) / 3,
		[self zone]);
	    
    while ((key = [keys nextObject]))
	NSMapInsert(table, key, [dictionary objectForKey:key]);
    
    return self;
}

- (void)dealloc
{
    if (table)
	NSFreeMapTable(table);
    [super dealloc];
}

/* Accessing keys and values */

- (NSEnumerator *)keyEnumerator
{
    return AUTORELEASE([[_NSConcreteDictionaryKeyEnumerator alloc]
                           initWithDictionary:self]);
}

- (id)objectForKey:(id)aKey
{
    return (NSObject*)NSMapGet(table, aKey);
}

- (unsigned int)count
{
    return NSCountMapTable(table);
}

/* Private */

- (NSMapEnumerator)__keyEnumerator;
{
    return NSEnumerateMapTable(table);
}

@end /* NSConcreteDictionary */

/*
 * NSConcreteEmptyDictionary class
 */

@implementation NSConcreteEmptyDictionary

/* Allocating and Initializing */

- (id)init
{
    return [self initWithDictionary:nil];
}

- (id)initWithObjects:(id*)objects
    forKeys:(id*)keys 
    count:(unsigned int)count
{
    if (count > 1) {
        RELEASE(self);
        return [[NSConcreteDictionary alloc]
                   initWithObjects:objects forKeys:keys count:count];
    }
    else if (count == 1) {
        RELEASE(self);
        return [[NSConcreteSingleObjectDictionary alloc]
                   initWithObjects:objects forKeys:keys count:count];
    }
    return self;
}

- (id)initWithDictionary:(NSDictionary*)dictionary
{
    unsigned count = [dictionary count];
    
    if (count > 1) {
        RELEASE(self);
        return [[NSConcreteDictionary alloc] initWithDictionary:dictionary];
    }
    else if (count == 1) {
        RELEASE(self);
        return [[NSConcreteSingleObjectDictionary alloc]
                                                  initWithDictionary:dictionary];
    }
    return self;
}

/* Accessing keys and values */

- (NSEnumerator *)keyEnumerator
{
    return AUTORELEASE([[_NSConcreteSingleObjectDictionaryKeyEnumerator alloc]
                            initWithObject:nil]);
}

- (id)objectForKey:(id)aKey
{
    return nil;
}

- (unsigned int)count
{
    return 0;
}

@end /* NSConcreteEmptyDictionary */

/*
 * NSConcreteSingleObjectDictionary class
 */

@implementation NSConcreteSingleObjectDictionary

+ (void)initialize {
    if (sharedEmptyDict == nil)
        sharedEmptyDict = [[NSConcreteEmptyDictionary alloc] init];
}

/* Allocating and Initializing */

- (id)init
{
    RELEASE(self);
    return RETAIN(sharedEmptyDict);
}

- (id)initWithObjects:(id*)objects
    forKeys:(id*)keys 
    count:(unsigned int)count
{
    if (count == 0) {
        RELEASE(self);
        return RETAIN(sharedEmptyDict);
    }
    else if (count > 1) {
        RELEASE(self);
        return [[NSConcreteDictionary alloc]
                   initWithObjects:objects forKeys:keys count:count];
    }

    if (*keys == nil) {
        THROW([[InvalidArgumentException alloc] 
		    initWithReason:@"Nil key to be added in dictionary"]);
    }
    if (*objects == nil) {
        THROW([[InvalidArgumentException alloc] 
		    initWithReason:@"Nil object to be added in dictionary"]);
    }
    self->key   = RETAIN(*keys);
    self->value = RETAIN(*objects);
    return self;
}

- (id)initWithDictionary:(NSDictionary*)dictionary
{
    unsigned count = [dictionary count];

    if (count == 0) {
        RELEASE(self);
        return RETAIN(sharedEmptyDict);
    }
    else if (count > 1) {
        RELEASE(self);
        return [[NSConcreteDictionary alloc]
                   initWithDictionary:dictionary];
    }

    self->key   = RETAIN([[dictionary keyEnumerator] nextObject]);
    self->value = RETAIN([dictionary objectForKey:key]);
    return self;
}

- (void)dealloc
{
    RELEASE(self->key);
    RELEASE(self->value);
    [super dealloc];
}

/* Accessing keys and values */

- (NSEnumerator *)keyEnumerator
{
    return AUTORELEASE([[_NSConcreteSingleObjectDictionaryKeyEnumerator alloc]
                            initWithObject:self->key]);
}

- (id)objectForKey:(id)aKey
{
    return ([self->key isEqual:aKey]) ? self->value : nil;
}

- (unsigned int)count
{
    return 1;
}

@end /* NSConcreteSingleObjectDictionary */

/*
 * NSConcreteSmallDictionary class
 */

@implementation NSConcreteSmallDictionary

- (id)init
{
    RELEASE(self);
    return RETAIN(sharedEmptyDict);
}

- (id)initWithObjects:(id*)_objects forKeys:(id*)_keys 
  count:(unsigned int)_count
{
    if (count == 0) {
        RELEASE(self);
        return RETAIN(sharedEmptyDict);
    }
    else if (count == 1) {
        RELEASE(self);
        return [[NSConcreteSingleObjectDictionary alloc]
                   initWithObjects:_objects forKeys:_keys count:_count];
    }
    
    NSAssert2(_count <= SMALL_NSDICTIONARY_SIZE,
              @"too many objects for small dictionary (max=%i, count=%i)!",
              SMALL_NSDICTIONARY_SIZE, _count);
    self->count = _count;

    while(_count--) {
	if (!_keys[_count] || !_objects[_count]) {
	    THROW([[InvalidArgumentException alloc] 
		    initWithReason:@"Nil object to be added in dictionary"]);
        }
        self->entries[_count].key   = RETAIN(_keys[_count]);
        self->entries[_count].hash  = [_keys[_count] hash];
        self->entries[_count].value = RETAIN(_objects[_count]);
    }
    return self;
}
- (id)initWithDictionary:(NSDictionary*)dictionary
{
    NSEnumerator  *keys = [dictionary keyEnumerator];
    unsigned char i;
    
    self->count = [dictionary count];
    NSAssert2(self->count <= SMALL_NSDICTIONARY_SIZE,
              @"too many objects for small dictionary (max=%i, count=%i)!",
              SMALL_NSDICTIONARY_SIZE, self->count);

    for (i = 0; i < self->count; i++) {
        id key = [keys nextObject];
        self->entries[i].key   = RETAIN(key);
        self->entries[i].hash  = [key hash];
        self->entries[i].value = RETAIN([dictionary objectForKey:key]);
    }
    return self;
}

- (void)dealloc {
    unsigned char i;
    for (i = 0; i < self->count; i++) {
        RELEASE(self->entries[i].key);
        RELEASE(self->entries[i].value);
    }
    [super dealloc];
}

- (id)objectForKey:(id)aKey
{
    register NSSmallDictionaryEntry *e = self->entries;
    register char i;
    register unsigned hash = [aKey hash];
    
    for (i = (self->count - 1); i >= 0; i--, e++) {
        if (e->hash == hash) {
            if (e->key == aKey)        return e->value;
            if ([e->key isEqual:aKey]) return e->value;
        }
    }
    return nil;
}

- (unsigned int)count
{
    return self->count;
}
- (NSEnumerator *)keyEnumerator
{
    return AUTORELEASE([[_NSConcreteSmallDictionaryKeyEnumerator alloc]
                            initWithDictionary:self
                            firstEntry:self->entries count:self->count]);
}

@end /* NSConcreteSmallDictionary */

/*
* NSConcreteMutableDictionary class
*/

@implementation NSConcreteMutableDictionary

+ (void)initialize
{
    static BOOL initialized = NO;
    if(!initialized) {
	initialized = YES;
	class_add_behavior(self, [NSConcreteDictionary class]);
    }
}

/* Allocating and Initializing */

- (id)init
{
    return [self initWithCapacity:0];
}

- (id)initWithCapacity:(unsigned int)aNumItems
{
    table = NSCreateMapTableWithZone(NSObjectMapKeyCallBacks,
		NSObjectMapValueCallBacks, 
		(aNumItems * 4) / 3,
		[self zone]);
    return self;
}

- (id)initWithObjects:(id*)objects
    forKeys:(id*)keys 
    count:(unsigned int)count
{
    table = NSCreateMapTableWithZone(NSObjectMapKeyCallBacks,
		NSObjectMapValueCallBacks, 
		(count * 4) / 3,
		[self zone]);

    if (count == 0) return self;

    while(count--) {
	if (!keys[count] || !objects[count])
	    THROW([[InvalidArgumentException alloc] 
		    initWithReason:@"Nil object to be added in dictionary"]);
	NSMapInsert(table, keys[count], objects[count]);
    }
    return self;
}

- (id)initWithDictionary:(NSDictionary*)dictionary
{
    id keys = [dictionary keyEnumerator];
    id key;

    table = NSCreateMapTableWithZone(NSObjectMapKeyCallBacks,
		NSObjectMapValueCallBacks, 
		([dictionary count] * 4) / 3,
		[self zone]);
	    
    while ((key = [keys nextObject]))
	NSMapInsert(table, key, [dictionary objectForKey:key]);

    return self;
}

- (void)dealloc
{
    if (table)
	NSFreeMapTable(table);
    [super dealloc];
}

/* Modifying dictionary */

- (void)setObject:(id)anObject forKey:(id)aKey
{
    if (!anObject || !aKey)
	THROW ([[InvalidArgumentException alloc]
		    initWithFormat:@"nil object to be added in dictionary"]);
    NSMapInsert(table, aKey, anObject);
}

- (void)removeObjectForKey:(id)aKey
{
    NSMapRemove(table, aKey);
}

- (void)removeAllObjects
{
    NSResetMapTable(table);
}

@end /* NSConcreteMutableDictionary */

/*
* NSConcreteDictionary Enumerator classes
*/

@implementation _NSConcreteDictionaryKeyEnumerator

- initWithDictionary:(NSDictionary*)_dict
{
    dict = RETAIN(_dict);
    enumerator = [(NSConcreteDictionary*)dict __keyEnumerator];
    return self;
}

- (void)dealloc
{
    RELEASE(dict);
    [super dealloc];
}

- nextObject
{
    id key, value;
    return NSNextMapEnumeratorPair(&enumerator, 
		(void**)&key, (void**)&value) == YES ? 
		key : nil;
}

@end /* _NSConcreteDictionaryKeyEnumerator */

@implementation _NSConcreteSingleObjectDictionaryKeyEnumerator

- (id)initWithObject:(id)_object
{
    self->nextObject = RETAIN(_object);
    return self;
}
- (void)dealloc {
    RELEASE(self->nextObject);
    [super dealloc];
}

- (id)nextObject
{
    id no = self->nextObject;
    if (no) {
        AUTORELEASE(self->nextObject);
        self->nextObject = nil;
    }
    return no;
}

@end

@implementation _NSConcreteSmallDictionaryKeyEnumerator

- (id)initWithDictionary:(NSConcreteSmallDictionary *)_dict
  firstEntry:(NSSmallDictionaryEntry *)_firstEntry
  count:(unsigned char)_count
{
    self->dict = RETAIN(_dict);
    self->currentEntry = _firstEntry;
    self->count        = _count;
    return self;
}

- (void)dealloc {
    RELEASE(self->dict);
    [super dealloc];
}

- (id)nextObject {
    if (self->count > 0) {
        id obj = self->currentEntry->value;
        self->currentEntry++;
        self->count--;
        return obj;
    }
    else
        return nil;
}

@end

/*
  Local Variables:
  c-basic-offset: 4
  tab-width: 8
  End:
*/

