/* 
   NSNotificationCenter.m

   Copyright (C) 1995, 1996 Ovidiu Predescu and Mircea Oancea.
   All rights reserved.

   Author: Mircea Oancea <mircea@jupiter.elcom.pub.ro>

   This file is part of libFoundation.

   Permission to use, copy, modify, and distribute this software and its
   documentation for any purpose and without fee is hereby granted, provided
   that the above copyright notice appear in all copies and that both that
   copyright notice and this permission notice appear in supporting
   documentation.

   We disclaim all warranties with regard to this software, including all
   implied warranties of merchantability and fitness, in no event shall
   we be liable for any special, indirect or consequential damages or any
   damages whatsoever resulting from loss of use, data or profits, whether in
   an action of contract, negligence or other tortious action, arising out of
   or in connection with the use or performance of this software.
*/

#include <Foundation/common.h>
#include <Foundation/NSNotification.h>
#include <Foundation/NSDictionary.h>
#include <Foundation/NSArray.h>
#include <Foundation/NSString.h>
#include <Foundation/NSUtilities.h>
#include <Foundation/NSThread.h>

#include <extensions/objc-runtime.h>
#include <extensions/NSException.h>
#include <extensions/NSException.h>
#include <extensions/GarbageCollector.h>
#include <extensions/exceptions/GeneralExceptions.h>

#include "PrivateThreadData.h"

#define DEFAULT_CAPACITY 1023

/*
 * Objects/selectors pair used in sets
 */

@interface NSNotificationListItem : NSObject
{
@public
    /* weak */ id  observer;	// observer that will receive selector
    SEL selector;	        // is a postNotification:
    NSNotificationListItem* next; // this is needed for keeping a
                                  // linked list of items to be removed
}
- initWithObject:(id)anObserver selector:(SEL)aSelector;
- (BOOL)isEqual:other;
- (unsigned)hash;
- (void)postNotification:(NSNotification*)notification;
@end

@implementation NSNotificationListItem

#if LIB_FOUNDATION_BOEHM_GC
+ (void)initialize
{
  class_ivar_set_gcinvisible (self, "observer", YES);
}
#endif

- initWithObject:(id)anObserver selector:(SEL)aSelector
{
    observer = anObserver;
    selector = aSelector;
    return self;
}

- (BOOL)isEqual:other
{
    if ([other isKindOfClass:[NSNotificationListItem class]]) {
	NSNotificationListItem* obj = other;
	return (observer == obj->observer) && SEL_EQ(selector, obj->selector);
    }
    
    return NO;
}

- (unsigned)hash
{
    return ((long)observer >> 4) + 
	   __NSHashCString(NULL, sel_get_name(selector));
}

- (void)postNotification:(NSNotification*)notification
{
    [observer performSelector:selector withObject:notification];
}

@end /* NSNotificationListItem */

/*
 * Register for objects to observer mapping
 */

@interface NSNotificationObserverRegister : NSObject
{
    NSHashTable* observerItems;
}
- init;
- (unsigned int)count;
- (void)addObjectsToList:(NSMutableArray*)list;
- (void)addObserver:observer selector:(SEL)selector;
- (void)removeObserver:observer;
@end

@implementation NSNotificationObserverRegister

- init
{
    observerItems = NSCreateHashTable(NSObjectHashCallBacks, DEFAULT_CAPACITY);
    return self;
}

- (void)dealloc
{
    NSFreeHashTable(observerItems);
    [super dealloc];
}

- (unsigned int)count
{
    return NSCountHashTable(observerItems);
}

- (void)addObjectsToList:(NSMutableArray*)list
{
    id reg;
    NSHashEnumerator itemsEnum = NSEnumerateHashTable(observerItems);

    while((reg = (id)NSNextHashEnumeratorItem(&itemsEnum)))
	[list addObject:reg];
}

- (void)addObserver:observer selector:(SEL)selector
{
    NSNotificationListItem* reg
        = AUTORELEASE([[NSNotificationListItem alloc]
                          initWithObject:observer selector:selector]);
    
    NSHashInsertIfAbsent(observerItems, reg);
}

- (void)removeObserver:observer
{
    NSNotificationListItem* listItem = nil;
    NSNotificationListItem* reg;
    NSHashEnumerator itemsEnum;
    
    itemsEnum = NSEnumerateHashTable(observerItems);
    
    while((reg = (id)NSNextHashEnumeratorItem(&itemsEnum)))
      if (reg->observer == observer) {
	/* Add 'reg' to the linked list of ListItems. We use this schema here
	   to avoid allocating anything, this can trigger finalization calls
	   in the case of Boehm's GC. */
	reg->next = listItem;
	listItem = reg;
      }
    
    while (listItem) {
	NSHashRemove(observerItems, listItem);
	listItem = listItem->next;
    }
}

@end /* NSNotificationObserverRegister */

/*
 * Register for objects to observer mapping
 */

@interface NSNotificationObjectRegister : NSObject
{
    NSMapTable* objectObservers;
    NSNotificationObserverRegister* nilObjectObservers;
}

- init;
- listToNotifyForObject:object;
- (void)addObserver:observer selector:(SEL)selector object:object;
- (void)removeObserver:observer object:object;
- (void)removeObserver:observer;

#if LIB_FOUNDATION_BOEHM_GC
- (void)removeObject:object;
#endif

@end

@implementation NSNotificationObjectRegister

- init
{
#if LIB_FOUNDATION_BOEHM_GC
    objectObservers = NSCreateMapTableInvisibleKeysOrValues
	(NSNonRetainedObjectMapKeyCallBacks,
	 NSNonRetainedObjectMapValueCallBacks,
	 DEFAULT_CAPACITY,
	 YES, NO);
#else
    objectObservers = NSCreateMapTable (NSNonRetainedObjectMapKeyCallBacks,
                                        NSObjectMapValueCallBacks,
                                        DEFAULT_CAPACITY);
#endif
    nilObjectObservers = [NSNotificationObserverRegister new];
    return self;
}

- (void)dealloc
{
    NSFreeMapTable(objectObservers);
    RELEASE(nilObjectObservers);
    [super dealloc];
}

- listToNotifyForObject:object
{
    id reg = nil;
    int count;
    id list;
    
    if (object)
	reg = (id)NSMapGet(objectObservers, object);
    count = [reg count] + [nilObjectObservers count];
    list = AUTORELEASE([[NSMutableArray alloc] initWithCapacity:count]);
    
    [reg addObjectsToList:list];
    [nilObjectObservers addObjectsToList:list];
    
    return list;
}

- (void)addObserver:observer selector:(SEL)selector object:object
{
    NSNotificationObserverRegister* reg;
    
    if (object) {
	reg = (id)NSMapGet(objectObservers, object);
	if (!reg) {
	    reg = AUTORELEASE([[NSNotificationObserverRegister alloc] init]);
	    NSMapInsert(objectObservers, object, reg);
	}
    }
    else
	reg = nilObjectObservers;
    
    [reg addObserver:observer selector:selector];
}

- (void)removeObserver:observer object:object
{
    NSNotificationObserverRegister* reg;
    
    if (object)
	reg = NSMapGet(objectObservers, object);
    else
	reg = nilObjectObservers;
    
    [reg removeObserver:observer];
}

- (void)removeObserver:observer
{
    id obj, reg;
    NSMapEnumerator regEnum = NSEnumerateMapTable(objectObservers);

    while (NSNextMapEnumeratorPair(&regEnum, (void*)&obj, (void*)&reg))
	[reg removeObserver:observer];

    [nilObjectObservers removeObserver:observer];
}

#if LIB_FOUNDATION_BOEHM_GC
- (void)removeObject:(id)object
{
    NSMapRemove (objectObservers, object);
}
#endif

@end /* NSNotificationObjectRegister */

/*
 * NSNotificationCenter	
 */

static NSNotificationCenter *defaultCenter = nil;
static BOOL isMultiThreaded = NO;

@implementation NSNotificationCenter 

/* Class methods */

+ (void)initialize
{
    static BOOL initialized = NO;

    if (!initialized) {
	initialized = YES;
	defaultCenter = [self alloc];
	[defaultCenter init];
    }
}

+ (NSNotificationCenter*)defaultCenter
{
    if (isMultiThreaded)
	return [[[NSThread currentThread] _privateThreadData]
		    defaultNotificationCenter];
    else
	return defaultCenter;
}

+ (void)taskNowMultiThreaded:notification
{
    PrivateThreadData* threadData = [[NSThread currentThread]
						_privateThreadData];

    [threadData setDefaultNotificationCenter:defaultCenter];
    defaultCenter = nil;
    isMultiThreaded = YES;
}

/* Init/dealloc */

- (id)init
{
#if LIB_FOUNDATION_BOEHM_GC
    nameToObjects = NSCreateMapTable  (NSNonRetainedObjectMapKeyCallBacks,
                                       NSNonOwnedPointerMapValueCallBacks,
                                       119);
#else
    nameToObjects = NSCreateMapTable (NSObjectMapKeyCallBacks,
				      NSObjectMapValueCallBacks,
				      119);
#endif
    nullNameToObjects = [NSNotificationObjectRegister new];
    return self;
}

- (void)dealloc
{
    NSFreeMapTable (nameToObjects);
    RELEASE(nullNameToObjects);
    [super dealloc];
}

/* Register && post notifications */
    
- (void)postNotification:(NSNotification*)notification
{
    NSArray* fromName;
    NSArray* fromNull;

    NSNotificationObjectRegister* reg;
    
    id name = [notification notificationName];
    id object = [notification notificationObject];

    if (name == nil)
	THROW([[InvalidArgumentException alloc]
	    initWithFormat:@"`nil' notification name in postNotification:"]);
    
    // get objects to notify with registered notification name
    reg = NSMapGet (nameToObjects, name);
    fromName = [reg listToNotifyForObject:object];
    
    // get objects to notify with no notification name
    fromNull = [nullNameToObjects listToNotifyForObject:object];

    // send notifications
    [fromName makeObjectsPerform:@selector(postNotification:)
	withObject:notification];
    [fromNull makeObjectsPerform:@selector(postNotification:)
	withObject:notification];
}

- (void)addObserver:observer selector:(SEL)selector 
  name:(NSString*)notificationName object:object
{
    NSNotificationObjectRegister* reg;
    
    if (notificationName == nil)
	reg = nullNameToObjects;
    else {
	notificationName
            = AUTORELEASE([notificationName
                              copyWithZone:[notificationName zone]]);
	reg = NSMapGet (nameToObjects, notificationName);
	if (!reg) {
	    reg = AUTORELEASE([[NSNotificationObjectRegister alloc] init]);
	    NSMapInsert (nameToObjects, notificationName, reg);
	}
    }
    [reg addObserver:observer selector:selector object:object];

#if LIB_FOUNDATION_BOEHM_GC
    if (object)
        [GarbageCollector registerForFinalizationObserver:self
			  selector:@selector(_objectWillFinalize:)
			  object:object];

    [GarbageCollector registerForFinalizationObserver:self
		      selector:@selector(_observerWillFinalize:)
		      object:observer];
#endif
}

- (void)removeObserver:observer 
  name:(NSString*)notificationName object:object
{
    NSNotificationObjectRegister* reg;

    if (notificationName == nil)
	reg = nullNameToObjects;
    else
	reg = NSMapGet (nameToObjects, notificationName);
    [reg removeObserver:observer object:object];
}

- (void)removeObserver:observer
{
    id name;
    NSMapEnumerator enumerator;
    NSNotificationObjectRegister* reg;

    enumerator = NSEnumerateMapTable (nameToObjects);

    while (NSNextMapEnumeratorPair (&enumerator, (void*)&name, (void*)&reg))
	[reg removeObserver:observer];

    [nullNameToObjects removeObserver:observer];
}

#if LIB_FOUNDATION_BOEHM_GC
- (void)removeObject:object
{
    id name;
    NSMapEnumerator enumerator;
    NSNotificationObjectRegister* reg;

    enumerator = NSEnumerateMapTable (nameToObjects);

    while (NSNextMapEnumeratorPair (&enumerator, (void*)&name, (void*)&reg))
	[reg removeObject:object];

    [nullNameToObjects removeObject:object];
}

- (void)_objectWillFinalize:(id)object
{
//  printf ("NSNotificationCenter _objectWillFinalize %lu\n", (unsigned long)object);

  [self removeObject:object];
}

- (void)_observerWillFinalize:(id)observer
{
//  printf ("NSNotificationCenter _observerWillFinalize: %lu\n", (unsigned long)observer);

  [self removeObserver:observer];
}
#endif

- (void)postNotificationName:(NSString*)notificationName object:object
{
    id notification;

    notification = [[NSNotification alloc] initWithName:notificationName
					   object:object
					   userInfo:nil];
    [self postNotification:notification];
    RELEASE(notification);
}

- (void)postNotificationName:(NSString*)notificationName object:object
  userInfo:(NSDictionary*)userInfo;
{
    id notification;

    notification = [[NSNotification alloc] initWithName:notificationName
					   object:object
					   userInfo:userInfo];
    [self postNotification:notification];
    RELEASE(notification);
}

@end /* NSNotificationCenter */

/*
  Local Variables:
  c-basic-offset: 4
  tab-width: 8
  End:
*/

