 /* 
   NSNumberFormatter.m

   Copyright (C) 1998 MDlink online service center, Helge Hess
   All rights reserved.

   Author: Martin Spindler (ms@mdlink.de), Helge Hess (helge@mdlink.de)

   This file is part of libFoundation.

   Permission to use, copy, modify, and distribute this software and its
   documentation for any purpose and without fee is hereby granted, provided
   that the above copyright notice appear in all copies and that both that
   copyright notice and this permission notice appear in supporting
   documentation.

   We disclaim all warranties with regard to this software, including all
   implied warranties of merchantability and fitness, in no event shall
   we be liable for any special, indirect or consequential damages or any
   damages whatsoever resulting from loss of use, data or profits, whether in
   an action of contract, negligence or other tortious action, arising out of
   or in connection with the use or performance of this software.
*/
// $Id: NSNumberFormatter.m,v 1.1 1998/12/30 21:35:06 helge Exp $

#include <ctype.h>
#include <Foundation/common.h>
#include <Foundation/exceptions/GeneralExceptions.h>
#include <Foundation/NSNumberFormatter.h>
#include <Foundation/NSValue.h>
#include <Foundation/NSString.h>

/*
  Missing:
  zero format strings
 */

#define posDefaultFormat @"0.00"
#define negDefaultFormat @"-0.00"

static void     _checkFormat(NSNumberFormatter *self, NSString *_format);
static unichar  separatorFromString(NSString *_string);

static NSString *_prefix(NSNumberFormatter *self, NSString *_format);
static NSString *_suffix(NSNumberFormatter *self, NSString *_format);
static NSString *cleanFormat(NSNumberFormatter *self, NSString *_format);
static NSString *digitsToDot(NSString *_string);
static NSString *digitsFromDot(NSString *_string, NSString *_sep);

static BOOL digitsForString(NSNumberFormatter *self,
                            NSString **digitStr, NSString *_string);

static BOOL checkMinMaxForNumber(NSNumberFormatter *self, NSNumber *_number);

@implementation NSNumberFormatter

- (id)init
{
    self->decimalSeparator  = '.';
    self->thousandSeparator = ',';
    return self;
}

- (void)dealloc
{
    RELEASE(self->negativeFormat);
    RELEASE(self->positiveFormat);
    RELEASE(self->maximum);
    RELEASE(self->minimum);
    [super dealloc];
}

// accessors

- (void)setFormat:(NSString *)_format
{
    NSArray *stringArray;

    //TODO: insert zeroFormat
    stringArray = [_format componentsSeparatedByString:@";"];

    switch ([stringArray count]) {
        case 1:
            [self setPositiveFormat:_format];
            [self setNegativeFormat:[@"-" stringByAppendingString:_format]];
            break;
        case 2:
            [self setPositiveFormat:[stringArray objectAtIndex:0]];
            [self setNegativeFormat:[stringArray objectAtIndex:1]];
            break;
        case 3:
            //TODO: insert zeroFormat (at index 1)
            NSLog(@"WARNING: zero format is not supported !");
            [self setPositiveFormat:[stringArray objectAtIndex:0]];
            [self setNegativeFormat:[stringArray objectAtIndex:2]];
            break;
        default:
            [[[InvalidArgumentException alloc]
                 initWithReason:@"invalid format passed to -setFormat:"] raise];
            break;
    }
}

- (NSString *)format
{
    NSMutableString *myString = nil;

    myString = [NSMutableString stringWithString:[self positiveFormat]];
    [myString appendString:@";"];
    [myString appendString:[self negativeFormat]];
    return myString;
}

- (void)setNegativeFormat:(NSString *)_format
{
    _checkFormat(self, _format);

    if (self->negativeFormat != _format) {
        RELEASE(self->negativeFormat);
        self->negativeFormat = [_format copyWithZone:[self zone]];
    }
}
- (NSString *)negativeFormat
{
    return self->negativeFormat ? self->negativeFormat : negDefaultFormat;
}

- (void)setPositiveFormat:(NSString *)_format
{
    _checkFormat(self, _format);

    if (self->positiveFormat != _format) {
        RELEASE(self->positiveFormat);
        self->positiveFormat = [_format copyWithZone:[self zone]];
    }
}
- (NSString *)positiveFormat
{
    return self->positiveFormat ? self->positiveFormat : posDefaultFormat;
}

- (void)setDecimalSeparator:(NSString *)_separator
{
    self->decimalSeparator = separatorFromString(_separator);
}
- (NSString *)decimalSeparator
{
    char c = self->decimalSeparator;
    return [NSString stringWithCString:&c length:1];
}

- (void)setThousandSeparator:(NSString *)_separator
{
    self->thousandSeparator = separatorFromString(_separator);
}
- (NSString *)thousandSeparator
{
    char c = self->thousandSeparator;
    return [NSString stringWithCString:&c length:1];
}

- (void)setMinimum:(NSNumber *)_number
{
    ASSIGN(self->minimum, _number);
}
- (NSNumber *)minimum
{
    return self->minimum;
}

- (void)setMaximum:(NSNumber *)_number
{
    ASSIGN(self->maximum, _number);
}
- (NSNumber *)maximum
{
    return self->maximum;
}

- (void)setHasThousandSeparators:(BOOL)_status
{
    self->hasThousandSeparator = _status;
}
- (BOOL)hasThousandSeparators
{
    return self->hasThousandSeparator;
}

- (void)setAllowsFloats:(BOOL)_flag
{
    [self notImplemented:_cmd];
}
- (BOOL)allowsFloats
{
    return YES;
}

// public methods

- (NSString *)stringForObjectValue:(id)anObject
{
    BOOL             hasThousandSep;
    NSString         *format;
    NSString         *numStr;
    NSString         *dummy;
    NSMutableString  *newString;
    int              len;          

    if (![anObject isKindOfClass:[NSNumber class]]) {
        [[[InvalidArgumentException alloc]
             initWithReason:@"number formatter formats only numbers !"] raise];
        return nil;
    }
    
    if ([anObject doubleValue] < 0.0) {
        // removes minus sign and chooses format (neg/pos)
        format = [self negativeFormat];
        numStr = [[anObject stringValue] substringFromIndex:1];
    }
    else {
        format = [self positiveFormat];
        numStr = [NSString stringWithString:[anObject stringValue]];
    }

    hasThousandSep = index([format cString], self->thousandSeparator) ? YES : NO;
    
    // removes all '#' and the 1000separators
    format = cleanFormat(self, format);
    NSAssert(format, @"invalid state !");
    
    //Digits in front of Decimal Separator
    newString = [digitsToDot(numStr) mutableCopy];
    AUTORELEASE(newString);

    if (self->hasThousandSeparator || hasThousandSep) {
        int length = [newString length];
        int i = 1;
        
        while (i*3 < length) {
            [newString insertString:[self thousandSeparator]
                       atIndex:(length - i * 3)];
            i++;
        }
    }
    len = [digitsToDot(format) length] - [digitsToDot(numStr) length];
    if (len > 0) {
        [newString insertString:[digitsToDot(format) substringToIndex:len]
                   atIndex:0];
    }
    [newString insertString:_prefix(self, format) atIndex:0];

    //Digits behind Decimal Separator
    if (index([format cString], self->decimalSeparator)) {
        [newString appendString:[self decimalSeparator]];

        dummy = digitsFromDot(format, [self decimalSeparator]);
        
        len = [dummy length] - [digitsFromDot(numStr, @".") length];
        if (len > 0) {
            [newString appendString:digitsFromDot(numStr, @".")];
            [newString appendString:
                         [dummy substringFromIndex:[dummy length] - len]];
        }
        else {
            [newString appendString:
                         [digitsFromDot(numStr, @".")
                                       substringToIndex:[dummy length]]];
        }
    }
    [newString appendString:_suffix(self, format)];
    
    return newString;
}


- (BOOL)getObjectValue:(id *)obj forString:(NSString *)string
  errorDescription:(NSString **)error
{
    NSString *num;
    BOOL     status;

    NSAssert(self->thousandSeparator != self->decimalSeparator,
             @"decimal and thousand separators are equal !");

    if (digitsForString(self, &num, string)) {
        status = YES;
        if (obj)   *obj   = [NSNumber numberWithDouble:[num doubleValue]];
        if (error) *error = @"";
    }
    else {
        status = NO;
        if (obj)   *obj   = [NSNumber numberWithDouble:[num doubleValue]];
        if (error) *error = @"wrong chars in number (only digits and sep's) !";
    }

    if (!checkMinMaxForNumber(self, *obj)) {
        status = NO;
        if (obj)   *obj   = nil;
        if (error) *error = @"number isn't in range of min and max! ";
    }
    return status;
}        

- (BOOL)isPartialStringValid:(NSString *)partialString
  newEditingString:(NSString **)newString
  errorDescription:(NSString **)error
{
    NSString *str     = nil;
    NSNumber *num     = nil;
    BOOL     status;

    *error = @"";

    //TODO: modify *newString (maybe it's not necessary)
    if (digitsForString(self, &str, partialString)) {
        status = YES;
    }
    else {
        status  = NO;
        if (error) *error = @"wrong chars in number (only digits and sep's) !";
        if (newString) *newString = str;
    }
    
    num = [NSNumber numberWithDouble:[str doubleValue]];
    if (!checkMinMaxForNumber(self, num)) {
        status  = NO;
        num     = nil;
        if (error) *error = @"number isn't in range of min and max! ";
    }
    return YES;
}


// Private Methods

static void _checkFormat(NSNumberFormatter *self, NSString *_format)
{
    char *format     = NULL;
    BOOL digitStatus = NO;
    int  decSepCount = 0; 
    int  i           = 0;

    NSCAssert(_format,               @"invalid state of format !");
    NSCAssert([_format length] != 0, @"empty format");
    
    format = MallocAtomic([_format cStringLength] + 1);
    
    strcpy(format, [_format cString]);
    while (format[i] != '\0') {
        if (format[i] == self->decimalSeparator) {
            digitStatus = NO;
            decSepCount++;         
        }
        if (index("0123456789_", format[i]) != NULL)
            digitStatus = YES;                    

        i++;
    }
    Free(format); format = NULL;
    
    NSCAssert(decSepCount <= 1, @"Too many decimal separators in format!");
    NSCAssert(digitStatus,
              @"format should be x or x.x (x is a digit or \'_\') !");
}

static unichar separatorFromString(NSString *_string)
{
   unichar theChar;

   NSCAssert(_string,              @"invalid separator parameter !");
   NSCAssert([_string length] > 0, @"separator has no chars !");
   
   if ([_string length] > 1)
       NSLog(@"WARNING: separator greater than 1 !");
   theChar = [_string characterAtIndex:0];

   NSCAssert(theChar < 256,
             @"unicode larger than 255, not supported (only 8-bit !)");
   NSCAssert(index("$0123456789#_", theChar) == NULL,
             @"invalid decimal separator char !");

   return theChar;
}

static NSString *cleanFormat(NSNumberFormatter *self, NSString *_format)
{
    const char *format = [_format cString];
    char *buffer = NULL;
    int  i,j;

    buffer = MallocAtomic([_format cStringLength] + 1);
    
    for (i = 0, j = 0; format[i] != '\0'; i++) {
        register char c = format[i];
        
        if ((c != self->thousandSeparator) && (c != '#')) {
            buffer[j] = format[i];
            j++;
        }
    }
    buffer[j] = '\0';
    format = NULL;
    
    return [NSString stringWithCStringNoCopy:buffer freeWhenDone:YES];
}


static NSString *_prefix(NSNumberFormatter *self, NSString *_format)
{
    const char *format = [_format cString];
    char *buffer = NULL;
    int  i = 0;

    buffer = MallocAtomic([_format cStringLength] + 1);

    while ((index("0123456789_", format[i]) == NULL) && (format[i] != '\0') &&
           (format[i] != self->decimalSeparator)) {
        buffer[i] = format[i];
        i++;
    }
    buffer[i] = '\0';
    format = NULL;
    
    return [NSString stringWithCStringNoCopy:buffer freeWhenDone:YES];
}

static NSString *_suffix(NSNumberFormatter *self, NSString *_format)
{
    char *buffer = NULL;
    char *format = NULL;
    int  i,j;

    format = MallocAtomic([_format cStringLength] +1);
    buffer = MallocAtomic([_format cStringLength] +1);
    [_format getCString:format];

    i = [_format cStringLength] - 1;
    j = 0;
    while ((index("0123456789_", format[i]) == NULL) && (i>=0) &&
           (format[i] !=self->decimalSeparator)) {
        buffer[j] = format[i];
        i--;
        j++;
    }
    buffer[j] = '\0';
    format[0] = '\0';

    i = strlen(buffer) - 1;
    j = 0;
    while (i >= 0) {
        format[j] = buffer[i];
        j++;
        i--;
    }
    format[j] = '\0';
    Free(buffer); buffer = NULL;
    
    return [NSString stringWithCStringNoCopy:format freeWhenDone:YES];
}

static NSString *digitsToDot(NSString *_string)
{
    const char *string = [_string cString];
    char *buffer = NULL;
    int  i = 0;
    int  j = 0;

    buffer = MallocAtomic([_string cStringLength] +1);

    while ((index("0123456789_", string[i]) == NULL) && (string[i] !='\0'))
        i++;

    while ((index("0123456789_", string[i]) != NULL) && (string[i] != '\0')) {
        buffer[j] = (string[i] == '_') ? ' ' : string[i];
        i++;
        j++;
    }
    buffer[j] = '\0';
    string = NULL;
    return [NSString stringWithCStringNoCopy:buffer freeWhenDone:YES];
}

static NSString *digitsFromDot(NSString *_string, NSString *_sep)
{
    char *buffer = NULL;
    char *string = NULL;
    char sep[2];
    int  i;
    int  j;

    if ([_string rangeOfString:_sep].length == 0)
        // if there is no decimal separator then return @""
        return @"";

    NSCAssert([_sep length] < 2, @"separator to long (max 1 char)");
    [_sep getCString:sep];

    string = MallocAtomic([_string cStringLength] +1);
    buffer = MallocAtomic([_string cStringLength] +1);
    [_string getCString:string];

    i = strlen(string)-1;
    while ((index("0123456789_", string[i]) == NULL) && (i >= 0))
        i--;
    
    j = 0;
    while ((index(sep, string[i]) == NULL) && (i >= 0)) {
        if (index("0123456789_", string[i]))
            buffer[j] = (string[i] == '_') ? ' ' : string[i];
        i--;
        j++;
    }
    if ((index(sep, string[i]) == NULL) && (i == 0))
        buffer[0] = '\0';
    buffer[j] = '\0';

    string[0] = '\0';

    i = strlen(buffer) - 1;
    j = 0;
    while (i>=0) {
        string[j] = buffer[i];
        j++;
        i--;
    }
    string[j] = '\0';
    Free(buffer); buffer = NULL;

    return [NSString stringWithCStringNoCopy:string freeWhenDone:YES];
}

static BOOL digitsForString(NSNumberFormatter *self,
                            NSString **digitStr, NSString *_string)
{
    const char *string = [_string cString];
    BOOL decStat = NO;
    BOOL status  = YES;
    char *buffer = NULL;
    int  i       = 0;
    int  j       = 0;
    int  k       = 0;

    buffer = MallocAtomic([_string cStringLength] +1);

    while ((index("0123456789", string[i]) == NULL) && (string[i] != '\0')) {
        if (string[i] == '-') {
            if (j == 0) {
                buffer[j] = string[i];
                j++;
            }
        }
        i++;
    }
    while (string[i] !='\0') {
        if (index("0123456789", string[i])) {
            buffer[j] = string[i];
            j++;
        }
        else if (string[i] == self->decimalSeparator) {
            if (!decStat) {
                decStat = YES;
                if (j == 0) {
                    buffer[j] = '0';
                    j++;
                }
                buffer[j] = string[i];
                j++;
            }
            else break;
        }
        else if (string[i] == self->thousandSeparator) {
        }
        else
            break;
        i++;
    }
    for (k = strlen(string) - 1;
         (index("0123456789_", string[k]) == NULL) && (k >= 0);
         k--)
        ;

    for(; i < k; i++) {
        status    = NO;
        buffer[j] = string[i];
        j++;
    }
    buffer[j] = '\0';
    string = NULL;

    *digitStr = [NSString stringWithCStringNoCopy:buffer freeWhenDone:YES];
    buffer = NULL;

    return status; 
}

static BOOL checkMinMaxForNumber(NSNumberFormatter *self, NSNumber *_number)
{
    if ((self->minimum) && ([_number compare:self->minimum] < 0))
        return NO;

    if ((self->maximum) && ([_number compare:self->maximum] > 0))
        return NO;

    return YES;
}

@end

/*
  Local Variables:
  c-basic-offset: 4
  tab-width: 8
  End:
*/
