/* 
   NSTask.m

   Copyright (C) 1995, 1996, 1997 Ovidiu Predescu and Mircea Oancea.
   All rights reserved.

   Author: Ovidiu Predescu <ovidiu@net-community.com>

   Based on the code written by Aleksandr Savostyanov <sav@conextions.com>.

   This file is part of libFoundation.

   Permission to use, copy, modify, and distribute this software and its
   documentation for any purpose and without fee is hereby granted, provided
   that the above copyright notice appear in all copies and that both that
   copyright notice and this permission notice appear in supporting
   documentation.

   We disclaim all warranties with regard to this software, including all
   implied warranties of merchantability and fitness, in no event shall
   we be liable for any special, indirect or consequential damages or any
   damages whatsoever resulting from loss of use, data or profits, whether in
   an action of contract, negligence or other tortious action, arising out of
   or in connection with the use or performance of this software.
*/

#include <config.h>
#include <extensions/objc-runtime.h>

#ifdef HAVE_LIBC_H
#  include <libc.h>
#else
#  include <unistd.h>
#endif

#if HAVE_VFORK_H
# include <vfork.h>
#endif

#include <errno.h>
#include <signal.h>

#include <sys/types.h>

#if HAVE_SYS_WAIT_H
# include <sys/wait.h>
# define WAIT_TYPE int
#else
  /* Old BSD union wait */
# define WAIT_TYPE union wait

# ifndef WEXITSTATUS
#  define WEXITSTATUS(stat_val) (int)(WIFEXITED(stat_val) \
					? (((stat_val.w_status) >> 8) & 0377) \
					: -1)
# endif
#endif

#ifndef WEXITSTATUS
# define WEXITSTATUS(stat_val) ((unsigned)(stat_val) >> 8)
#endif

#ifndef WIFEXITED
# define WIFEXITED(stat_val) (((stat_val) & 255) == 0)
#endif

#include <sys/resource.h>

#if HAVE_SYS_PARAM_H
# include <sys/param.h>
#endif

#if HAVE_SYS_TIME_H
# include <sys/time.h>	/* for struct timeval */
#endif

#include <Foundation/common.h>
#include <Foundation/NSTask.h>
#include <Foundation/NSRunLoop.h>
#include <Foundation/NSArray.h>
#include <Foundation/NSDictionary.h>
#include <Foundation/NSString.h>
#include <Foundation/NSValue.h>
#include <Foundation/NSAutoreleasePool.h>
#include <Foundation/NSDate.h>
#include <Foundation/NSFileHandle.h>
#include <Foundation/NSNotification.h>
#include <Foundation/NSException.h>
#include <Foundation/NSMapTable.h>
#include <Foundation/UnixSignalHandler.h>

#include <Foundation/exceptions/GeneralExceptions.h>


NSString *NSTaskDidTerminateNotification = @"NSTaskDidTerminateNotification";


@interface NSConcreteTask : NSTask
{
    pid_t        pid;
    NSString     *taskPath;
    NSString     *currentDirectory;
    NSArray      *taskArguments;
    NSDictionary *taskEnvironment;
    int          status;
    id           standardInput;  // either NSPipe or NSFileHandle
    id           standardOutput; // either NSPipe or NSFileHandle
    id           standardError;  // either NSPipe or NSFileHandle
    BOOL         isRunning;
    BOOL         childHasFinished;
}
@end

static NSMapTable* unixProcessToTask = NULL;


@implementation NSTask

+ (void)initialize
{
    static BOOL initialized = NO;

    if (!initialized) {
	initialized = YES;
	unixProcessToTask = NSCreateMapTable (NSIntMapKeyCallBacks,
				    NSNonRetainedObjectMapValueCallBacks, 19);
    }
}

+ (id)allocWithZone:(NSZone*)zone
{
    NSTask* task = NSAllocateObject((self == [NSTask class]) ? 
	    [NSConcreteTask class] : self, 0, zone);

    return task;
}

+ (NSTask*)launchedTaskWithLaunchPath:(NSString*)path
  arguments:(NSArray*)arguments
{
    id aTask;

    aTask = [[NSConcreteTask alloc] init];
    [aTask setLaunchPath:path];
    [aTask setArguments:arguments];
    [aTask launch];

    return AUTORELEASE(aTask);
}

- (void)setLaunchPath:(NSString*)path
{
    [self shouldNotImplement:_cmd];
}

- (void)setArguments:(NSArray*)arguments
{
    [self shouldNotImplement:_cmd];
}

- (void)setEnvironment:(NSDictionary*)dict
{
    [self shouldNotImplement:_cmd];
}

- (void)setCurrentDirectoryPath:(NSString*)path
{
    [self shouldNotImplement:_cmd];
}

- (void)setStandardInput:(id)input
{
    [self shouldNotImplement:_cmd];
}

- (void)setStandardOutput:(id)output
{
    [self shouldNotImplement:_cmd];
}

- (void)setStandardError:(id)error
{
    [self shouldNotImplement:_cmd];
}

- (NSString*)launchPath
{
    [self shouldNotImplement:_cmd];
    return nil;
}

- (NSArray*)arguments
{
    [self shouldNotImplement:_cmd];
    return nil;
}

- (NSDictionary*)environment
{
    [self shouldNotImplement:_cmd];
    return nil;
}

- (NSString*)currentDirectoryPath
{
    [self shouldNotImplement:_cmd];
    return nil;
}

- (id)standardInput
{
    [self shouldNotImplement:_cmd];
    return nil;
}

- (id)standardOutput
{
    [self shouldNotImplement:_cmd];
    return nil;
}

- (id)standardError
{
    [self shouldNotImplement:_cmd];
    return nil;
}

- (void)launch
{
    [self shouldNotImplement:_cmd];
}

- (void)terminate
{
    [self shouldNotImplement:_cmd];
}

- (BOOL)isRunning
{
    [self shouldNotImplement:_cmd];
    return NO;
}

- (int)terminationStatus
{
    [self shouldNotImplement:_cmd];
    return 0;
}

- (void)waitUntilExit
{
    [self shouldNotImplement:_cmd];
}

- (unsigned int)processId
{
    [self shouldNotImplement:_cmd];
    return 0;
}

@end


@implementation NSConcreteTask

+ (void)initialize
{
    static BOOL initialized = NO;

    if (!initialized) {
	initialized = YES;
	[[UnixSignalHandler sharedHandler]
	    addObserver:self
	    selector:@selector(_childFinished:)
#if defined(SIGCHLD)
	    forSignal:SIGCHLD
#elif defined(SIGCLD)
	    forSignal:SIGCLD
#endif
	    immediatelyNotifyOnSignal:YES];
    }
}

- (void)dealloc
{
    if (self->pid)
	NSMapRemove (unixProcessToTask, (void*)(long)pid);

    RELEASE(taskPath);
    RELEASE(currentDirectory);
    RELEASE(taskArguments);
    RELEASE(taskEnvironment);
    RELEASE(standardInput);
    RELEASE(standardOutput);
    RELEASE(standardError);
    [super dealloc];
}

- (void)setLaunchPath:(NSString*)path
{
    if (self->isRunning || self->childHasFinished) {
        [NSException raise:NSInvalidArgumentException
                     format:@"task is already launched."];
    }
    ASSIGN(taskPath, path);
}

- (void)setArguments:(NSArray*)arguments
{
    if (self->isRunning || self->childHasFinished) {
        [NSException raise:NSInvalidArgumentException
                     format:@"task is already launched."];
    }
    ASSIGN(taskArguments, arguments);
}

- (void)setEnvironment:(NSDictionary*)dict
{
    if (self->isRunning || self->childHasFinished) {
        [NSException raise:NSInvalidArgumentException
                     format:@"task is already launched."];
    }
    ASSIGN(taskEnvironment, dict);
}

- (void)setCurrentDirectoryPath:(NSString*)path
{
    if (self->isRunning || self->childHasFinished) {
        [NSException raise:NSInvalidArgumentException
                     format:@"task is already launched."];
    }
    ASSIGN(currentDirectory, path);
}

- (void)setStandardInput:(id)input
{
    if (self->isRunning || self->childHasFinished) {
        [NSException raise:NSInvalidArgumentException
                     format:@"task is already launched."];
    }
    ASSIGN(self->standardInput, input);
}

- (void)setStandardOutput:(id)output
{
    if (self->isRunning || self->childHasFinished) {
        [NSException raise:NSInvalidArgumentException
                     format:@"task is already launched."];
    }
    ASSIGN(self->standardOutput, output);
}

- (void)setStandardError:(id)error
{
    if (self->isRunning || self->childHasFinished) {
        [NSException raise:NSInvalidArgumentException
                     format:@"task is already launched."];
    }
    ASSIGN(self->standardError, error);
}

- (NSString*)launchPath
{
    return taskPath;
}
- (NSArray*)arguments
{
    return taskArguments;
}
- (NSDictionary*)environment
{
    return taskEnvironment;
}
- (NSString*)currentDirectoryPath
{
    return currentDirectory;
}

- (id)standardInput
{
    return self->standardInput;
}
- (id)standardOutput
{
    return self->standardOutput;
}
- (id)standardError
{
    return self->standardError;
}

- (BOOL)isRunning
{
    return self->isRunning;
}
- (unsigned int)processId
{
    return self->pid;
}

- (void)_execChild
{
    int          i, count, fd;
    char         *path, **parg, **penv;
    NSEnumerator *enumerator;

    if (self->standardInput) {
        int fd;
        
	close(0);

        fd = [self->standardInput isKindOfClass:[NSPipe class]]
            ? [[self->standardInput fileHandleForReading] fileDescriptor]
            : [self->standardInput fileDescriptor];
        dup2(fd, 0);
    }
    if (self->standardOutput) {
	close(1);
        fd = [self->standardOutput isKindOfClass:[NSPipe class]]
            ? [[self->standardOutput fileHandleForWriting] fileDescriptor]
            : [self->standardOutput fileDescriptor];
        dup2(fd, 1);
    }
    if (self->standardError) {
	close(2);
        fd = [self->standardError isKindOfClass:[NSPipe class]]
            ? [[self->standardError fileHandleForWriting] fileDescriptor]
            : [self->standardError fileDescriptor];
        dup2(fd, 2);
    }

    // close all descriptors but stdin, stdout and stderr (0,1 and 2)
    // (this close procedure includes the pipe descriptors !)
    for (fd = 3; fd < NOFILE; fd++)
	close(fd);

    if (self->currentDirectory)
	chdir([self->currentDirectory cString]);

    count   = [taskArguments count];
    parg    = (char**)Malloc ((count + 2) * sizeof(void*));
    parg[0] = Strdup ([taskPath cString]);
    for (i = 0; i < count; i++)
	parg[i + 1] = Strdup ([[taskArguments objectAtIndex:i] cString]);
    parg[count + 1] = NULL;

    count = [taskEnvironment count];
    penv = (char**)Malloc ((count + 1) * sizeof(void*));
    enumerator = [taskEnvironment keyEnumerator];
    for (i = 0; i < count; i++) {
	NSString* key = [enumerator nextObject];
	const char* keyCString = [key cString];
	const char* valueCString = [[taskEnvironment objectForKey:key]
					cString];
	char buffer[Strlen(keyCString) + Strlen(valueCString) + 2];

	sprintf (buffer, "%s=%s", keyCString, valueCString);
	penv[i] = Strdup (buffer);
    }
    penv[count] = NULL;

    path = Strdup ([taskPath cString]);
    if (execve (path, parg, penv) == -1) {
	NSLog(@"Can't launch the child process, exec() failed!",
	      strerror (errno));
    }
    Free(parg);
    Free(penv);
}

+ (void)_childFinished:(int)signum
{
    WAIT_TYPE s;
    pid_t unixPid;

    unixPid = wait (&s);
    if ((long)unixPid != -1) {
	NSConcreteTask* task =
            NSMapGet (unixProcessToTask, (void*)(long)unixPid);

	if (!task)
	    return;

	if (WIFEXITED(s))
	    task->status = WEXITSTATUS(s);

	task->isRunning = NO;
	task->childHasFinished = YES;
    }
}

- (void)launch
{
    if (!taskPath)
	THROW([[InvalidArgumentException alloc]
		  initWithReason:@"the task's executable name is not setup!"]);

    if (isRunning)
	return;

    status = 0;

#if defined(SIGCHLD)
    [[UnixSignalHandler sharedHandler] blockSignal:SIGCHLD];
#elif defined(SIGCLD)
    [[UnixSignalHandler sharedHandler] blockSignal:SIGCLD];
#endif

    switch (pid = vfork()) {
	case -1:	/* error */
	    NSLog(@"Can't launch the child process, vfork() failed: %s!",
		  strerror (errno));
	    break;

	case 0:
#if defined(SIGCHLD)
	    [[UnixSignalHandler sharedHandler] enableSignal:SIGCHLD];
#elif defined(SIGCLD)
	    [[UnixSignalHandler sharedHandler] enableSignal:SIGCLD];
#endif
	    [self _execChild];
	    break;

	default:
	    isRunning = YES;
	    NSMapInsert (unixProcessToTask, (void*)(long)pid, self);
#if defined(SIGCHLD)
	    [[UnixSignalHandler sharedHandler] enableSignal:SIGCHLD];
#elif defined(SIGCLD)
	    [[UnixSignalHandler sharedHandler] enableSignal:SIGCLD];
#endif
            // close handles of pipes
            if ([self->standardInput isKindOfClass:[NSPipe class]])
                [[self->standardInput fileHandleForReading] closeFile];
            else
                [self->standardInput closeFile];
            
            if ([self->standardOutput isKindOfClass:[NSPipe class]])
                [[self->standardOutput fileHandleForWriting] closeFile];
            else
                [self->standardOutput closeFile];
                
            if ([self->standardError isKindOfClass:[NSPipe class]])
                [[self->standardError fileHandleForWriting] closeFile];
            else
                [self->standardError closeFile];

	    break;
    }
}

- (void)terminate
{
    if (childHasFinished) {
	/* -terminate has been already sent and the child finished */
	return;
    }

    if (!isRunning)
	THROW ([[InvalidArgumentException alloc]
		    initWithReason:@"task has not been launched yet!"]);

    if (pid)
	kill(pid, SIGTERM);

    /* Post the termination notification. A better idea would be to post this
       notification after the child has exited, by adding the task object
       as an observer to UnixSignalHandler in the signal handler function.
       But we keep here the same semantics with that in documentation. */
    [[NSNotificationCenter defaultCenter]
	postNotification:
	    [NSNotification notificationWithName:NSTaskDidTerminateNotification
			    object:self]];
}

- (int)terminationStatus
{
    if (isRunning)
	THROW([[InvalidArgumentException alloc]
		  initWithReason:@"task is still running!"]);
    return status;
}

- (void)waitUntilExit
{
    NSRunLoop *runLoop = [NSRunLoop currentRunLoop];

    while(isRunning) {
	NSDate *aDate;
	CREATE_AUTORELEASE_POOL(pool);

	aDate = [runLoop limitDateForMode:NSDefaultRunLoopMode];
	[runLoop acceptInputForMode:NSDefaultRunLoopMode beforeDate:aDate];
	RELEASE(pool);
    }
}

// description

- (NSString *)description
{
    /* Don't use -[NSString stringWithFormat:] method because it can cause
       infinite recursion. */
    char buffer[512];

    sprintf(buffer,
            "<0x%08X<%s> isRunning=%s childHasFinished=%s pid=%d>",
            (unsigned)self, (char*)object_get_class_name(self),
            self->isRunning ? "YES" : "NO",
            self->childHasFinished ? "YES" : "NO",
            (unsigned)self->pid);
    return [NSString stringWithCString:buffer];
}

@end
/*
  Local Variables:
  c-basic-offset: 4
  tab-width: 8
  End:
*/
